#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/wait.h>



/* *** Leider ist diese Methode der Authentication bereits mit   ***
   *** Multiple-Sessions-Angriff geknackt worden!                ***

   Wie kann man sicher berprfen, ob die Gegenseite die ist fr die
   sie sich ausgibt? Mit dieser Methode:

   1.) Beide Seiten erzeugen einen Zufallsstring
   2.) gegenseitiges zusenden dieses Strings
   3.) an diesen String den Authstring anhngen
   4.) davon die MD5Summe bilden
   5.) empfangen des Zufallsstring von der Gegenseite
   6.) an diesen String den Authstring anhngen
   7.) davon die MD5Summe bilden
   8.) Prfsummen austauschen
   9.) Prfen, ob die von der Gegenseite erzeugte Prfsumme mit der
       eigenen bereinstimmt, wenn Ja alles OK

   Das gute daran, der Authstring wird nicht bertragen, also kann
   diesen auch nicht geklaut werden. Da man aus der MD5Sums den
   Originalstring nicht mehr bilden kann, kann diese ohne bedenken
   bermittelt werden...

   Genau das tut diese Funktion:

   if (SockMD5Auth(socket des gegenber, authstring) == 1) {
     Alles OK
   }

*/


int SockMD5Auth(int fd, char *pwd)
{
  FILE *random, *f;
  unsigned int seed;
  int i;
  int sizerstr0, sizerstr1;
  char rstr0[256], rstr1[256];
  char rstr0sum[35], rstr1sum[35];

  int sockets_in[2];
  int sockets_out[2];

  // open /dev/urandom
  if ((random = fopen("/dev/urandom", "r")) == NULL) {
    random = fopen("/dev/random", "r");
  }
  // Init C-Rand.
  fread(&seed, 1, 4, random);
  srand(seed);
  // Size fr Auth-String
  sizerstr0 = 64 + (int) (60.0 * rand() / (RAND_MAX + 1.0));
  // Die erstern 32 Byte mit RAND fllen...
  for (i = 0; i < 32; i++) {
    rstr0[i] = (int) (256.0 * rand() / (RAND_MAX + 1.0));
  }
  // Den Rest von urandom
  fread(rstr0 + 32, sizerstr0 - 32, 1, random);
  // keine Nullen im String, Null am Ende
  for (i = 0; i < sizerstr0; i++) {
    if (rstr0[i] == 0) {
      rstr0[i] = 54;
    }
  }
  rstr0[sizerstr0] = 0;

  // Socketpair zur Kommunikation mit den froks erzeugen
  /* Get the socket pair */
  if (socketpair(AF_UNIX, SOCK_STREAM, 0, sockets_in) < 0) {
    printf("error %d on socketpair\n", errno);
    exit(1);
  }
  /* Get the socket pair */
  if (socketpair(AF_UNIX, SOCK_STREAM, 0, sockets_out) < 0) {
    printf("error %d on socketpair\n", errno);
    exit(1);
  }

  // fork zum erzeugen der Prfsumme
  if (!fork()) {
    close(sockets_in[1]);
    close(sockets_out[1]);
    /*    close(fd);
    if (f != NULL) {
      fclose(f);
      }*/

    if (fcntl(sockets_in[0], F_SETFD, 0) == -1) {
      perror("fcntl(..., F_SETFD, 0);");
    }
    if (fcntl(sockets_out[0], F_SETFD, 0) == -1) {
      perror("fcntl(..., F_SETFD, 0);");
    }

    if (dup2(sockets_in[0], 0) == -1) {
      perror("dup2();");
    }
    if (dup2(sockets_out[0], 1) == -1) {
      perror("dup2();");
    }
    if (dup2(sockets_out[0], 2) == -1) {
      perror("dup2();");
    }

    system("md5sum");

    close(sockets_in[0]);
    close(sockets_out[0]);
    close(0);
    close(1);
    close(2);

    exit(0);
  }

  close(sockets_in[0]);
  close(sockets_out[0]);

  // for zum schreiben der Daten an fork 0 bzw. md5sum
  if (!fork()) {
    close(sockets_out[1]);
    /*    close(fd);
    if (f != NULL) {
      fclose(f);
      }*/
    write(sockets_in[1], rstr0, sizerstr0);
    write(sockets_in[1], pwd, strlen(pwd));
    close(sockets_in[1]);

    exit(0);
  }
  close(sockets_in[1]);

  // lesen der Prfsumme
  read(sockets_out[1], rstr0sum, 32);
  close(sockets_out[1]);
  rstr0sum[32] = 0;

  // Sende den Zufallsstring
  SockSendString(fd, rstr0);

  // Gegenber
  // Empfangen des Zufallsstrings
  sizerstr1 = SockRecvString(fd, rstr1);

  // Socketpair zur Kommunikation mit den froks erzeugen
  /* Get the socket pair */
  if (socketpair(AF_UNIX, SOCK_STREAM, 0, sockets_in) < 0) {
    printf("error %d on socketpair\n", errno);
    exit(1);
  }
  /* Get the socket pair */
  if (socketpair(AF_UNIX, SOCK_STREAM, 0, sockets_out) < 0) {
    printf("error %d on socketpair\n", errno);
    exit(1);
  }

  // fork zum erzeugen der Prfsumme
  if (!fork()) {
    close(sockets_in[1]);
    close(sockets_out[1]);
    /*    close(fd);
    if (f != NULL) {
      fclose(f);
      }*/

    if (fcntl(sockets_in[0], F_SETFD, 0) == -1) {
      perror("fcntl(..., F_SETFD, 0);");
    }
    if (fcntl(sockets_out[0], F_SETFD, 0) == -1) {
      perror("fcntl(..., F_SETFD, 0);");
    }

    if (dup2(sockets_in[0], 0) == -1) {
      perror("dup2();");
    }
    if (dup2(sockets_out[0], 1) == -1) {
      perror("dup2();");
    }
    if (dup2(sockets_out[0], 2) == -1) {
      perror("dup2();");
    }

    system("md5sum");

    close(sockets_in[0]);
    close(sockets_out[0]);
    close(0);
    close(1);
    close(2);

    exit(0);
  }

  close(sockets_in[0]);
  close(sockets_out[0]);

  // for zum schreiben der Daten an fork 0 bzw. md5sum
  if (!fork()) {
    close(sockets_out[1]);
    /*    close(fd);
    if (f != NULL) {
      fclose(f);
      }*/
    write(sockets_in[1], rstr1, sizerstr1);
    write(sockets_in[1], pwd, strlen(pwd));
    close(sockets_in[1]);

    exit(0);
  }
  close(sockets_in[1]);

  // lesen der Prfsumme
  read(sockets_out[1], rstr1sum, 32);
  close(sockets_out[1]);
  rstr1sum[32] = 0;

  // Senden der aus dem Zufallsstring ermittelte Summe
  SockSendString(fd, rstr1sum);

  // Empfangen der aus dem Zufallsstrings ermittelte Summe
  SockRecvString(fd, rstr1sum);

  // Leichen beseitigen
  while(waitpid(-1,NULL,WNOHANG) > 0);

  // Vergleichen
  if (strcmp(rstr0sum, rstr1sum) == 0) {
    // OK
    SockSendString(fd, "OK");
    SockRecvString(fd, rstr0sum);
    if (strcmp(rstr0sum, "OK") == 0) {
      return 1;
    } else {
      return 0;
    }
  } else {
    // ERROR
    SockSendString(fd, "ERROR");
    return 0;
  }
}
